"""
The :mod:`pygslib.statistics` module includes statistics functions.
"""
from __future__ import absolute_import, division, print_function

from .kde import *
from .cdf import *
from .utils import *
from .postsim  import postsim_multfiles
